@@ -3,15 +3,18 @@
 #  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
 #  ------------------------------------------------------------------------------------------
 
-PANDA_DATASET_ID = "PANDA"
+PANDA_DATASET_ID = "panda"
 PANDA_5X_TILES_DATASET_ID = "PANDA_5X_tiles_20220714_134233_level1_224"
 PANDA_20X_TILES_DATASET_ID = "PANDA_20X_tiles_20220714_142937_level0_224"
 
 TCGA_CRCK_DATASET_ID = "TCGA-CRCk"
-TCGA_PRAD_DATASET_ID = "TCGA-PRAD"
+
+TCGA_PRAD_DATASET_ID = "TCGA-PRAD_20220712"
+TCGA_PRAD_10X_TILES_DATASET_ID = "TCGA-PRAD_10X_tiles_20220728_123811_level1_224"
 
 DEFAULT_DATASET_LOCATION = "/tmp/datasets/"
 PANDA_5X_TILES_DATASET_DIR = DEFAULT_DATASET_LOCATION + PANDA_5X_TILES_DATASET_ID
 PANDA_20X_TILES_DATASET_DIR = DEFAULT_DATASET_LOCATION + PANDA_20X_TILES_DATASET_ID
 TCGA_CRCK_DATASET_DIR = DEFAULT_DATASET_LOCATION + TCGA_CRCK_DATASET_ID
 TCGA_PRAD_DATASET_DIR = DEFAULT_DATASET_LOCATION + TCGA_PRAD_DATASET_ID
+TCGA_PRAD_10X_TILES_DATASET_DIR = DEFAULT_DATASET_LOCATION + TCGA_PRAD_DATASET_ID
