@@ -13,13 +13,12 @@ from typing import Any, Dict, Iterable, Optional, Tuple, Union
 import numpy as np
 import PIL
 from monai.data import Dataset
-from monai.data.image_reader import WSIReader
 from tqdm import tqdm
 from health_ml.utils.box_utils import Box
 
 from health_cpath.datasets.base_dataset import SlidesDataset
 from health_cpath.preprocessing import tiling
-from health_cpath.preprocessing.loading import LoadROId, segment_foreground
+from health_cpath.preprocessing.loading import LoadROId, WSIBackend, segment_foreground
 from health_cpath.utils.naming import SlideKey, TileKey
 
 
@@ -177,7 +176,7 @@ def process_slide(sample: Dict[SlideKey, Any], level: int, margin: int, tile_siz
             failed_tiles_file.write('tile_id' + '\n')
 
             print(f"Loading slide {slide_id} ...")
-            loader = LoadROId(WSIReader('cuCIM'), level=level, margin=margin,
+            loader = LoadROId(backend=WSIBackend.CUCIM, level=level, margin=margin,
                               foreground_threshold=foreground_threshold)
             sample = loader(sample)  # load 'image' from disk
 
